"""
########################################################################
The soil_water.py module contains the SoilWaterSeries class, which
provides I/O tools for processing measured volumetric soil water content
(SWC, cm3/cm3) data in the pyfao56 environment. The SoilWaterSeries
class manages SWC data collected at one location over time (e.g., at one
access tube over a growing season). A subclass called SoilWaterProfile
handles data storage and computations for one soil water profile
measurement event (i.e., measurements of the soil water profile on a
single date). The SoilWaterProfile class computes root zone soil water
metrics, especially the root zone soil water depletion (SWD, mm), given
estimates of root depth.

The soil_water.py module contains the following:
    SoilWaterSeries - A class for managing measured SWC profile series
    SoilWaterProfile - A class for managing a single SWC profile

10/17/2022 SWC Python functions developed by Josh Brekel, USDA-ARS
03/07/2023 SoilWater functions developed by Josh Brekel, USDA-ARS
08/23/2023 Major overhaul for pyfao56 1.2 release
########################################################################
"""

import pandas as pd
import datetime

class SoilWaterSeries:
    """A class for managing a series of measured soil water content data

    Attributes
    ----------
    par : pyfao56 Parameters object, optional
        Provides the parameter data (e.g., thetaFC, Zrmax)
        (default = None)
    sol : pyfao56 SoilProfile object, optional
        Provides layered soil profile data (e.g., thetaFC)
        (default = None)
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    swdata : dict
        Container for SoilWaterProfile objects
        key - measurement date as string ('yyyy-ddd')
        value - SoilWaterProfile object

    Methods
    -------
    savefile(filepath='pyfao56.sws')
        Save soil water series data to a file
    loadfile(filepath='pyfao56.sws')
        Load soil water series data from a file
    addprofile(mdate,swp)
        Add a SoilWaterProfile object to self.swdata
    customload()
        Override this function to customize loading measured volumetric
        soil water content data.
    summarize()
        Summarize the series of root zone soil water metrics
    """

    def __init__(self,filepath=None,par=None,sol=None,comment=''):
        """Initialize the SoilWaterSeries class attributes.

        If filepath is provided, soil water data is loaded from the file

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = None)
        par : pyfao56 Parameters object, optional
            Provides the parameter data (e.g., thetaFC, Zrmax)
            (default = None)
        sol : pyfao56 SoilProfile object, optional
            Provides layered soil profile data (e.g., thetaFC)
            (default = None)
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.par = par
        self.sol = sol
        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.swdata = {}

        if filepath is not None:
            self.loadfile(filepath)

    def __str__(self):
        """Represent the SoilWaterSeries class data as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        ast='*'*72
        s=('{:s}\n'
           'pyfao56: FAO-56 Evapotranspiration in Python\n'
           'Measured Soil Water Data\n'
           'Timestamp: {:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:s}\n'
          ).format(ast,timestamp,ast,self.comment,ast)
        if len(self.swdata) > 0:
            s += 'Year-DOY  n'
            key0 = list(self.swdata.keys())[0]
            n = len(self.swdata[key0].mvswc)
            for i in range(n):
                s += ' D{:02d}'.format(i+1)
            for i in range(n):
                s += ' SWC{:02d}'.format(i+1)
            s += '    Zr      mDr   mDrmax   mfDr mfDrmax mSWCr mSWCrmax'
            s += '    mKs\n'
            for key in sorted(self.swdata.keys()):
                s += self.swdata[key].__str__() + '\n'
        return s

    def savefile(self,filepath='pyfao56.sws'):
        """Save pyfao56 soil water series data to a file.

        The function saves a file with standard pyfao56-styled header.
        Each data line contains the following info:
        1. SWC measurement date as string ('yyyy-ddd')
        2. The number, x, of SWC measurement depths as integer
        3. x bottom depths of SWC measurement layers as integer (cm)
        4. x SWC measurements as integer (cm3/cm3, same order as depths)
        5. Estimate of Zr on the measurement date
        6. mDr (mm)
        7. mDrmax (mm)
        8. mfDr (mm/mm)
        9. mfDrmax (mm/mm)
        10. mSWCr (cm3/cm3)
        11. mSWCrmax (cm3/cm3)
        12. mKs (0.0-1.0)

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.sws')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for soil water data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.sws'):
        """Load measured soil water content data from a file

        The function expects a file with standard pyfao56-styled header.
        Each data line contains the following info:
        1. SWC measurement date as string ('yyyy-ddd')
        2. The number, x, of SWC measurement depths as integer
        3. x bottom depths of SWC measurement layers as integer (cm)
        4. x SWC measurements as integer (cm3/cm3, same order as depths)
        5. Optionally, an estimate of Zr on the measurement date
        6. Optionally, other values, which are not loaded.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.sws')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for soil water data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1]
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            self.swdata.clear()
            for line in lines[endast+2:]:
                line = line.strip().split()
                mdate = line[0]
                numdpths = int(line[1])
                mvswc = dict()
                for i in list(range(numdpths)):
                    dpth = int(line[2+i])
                    swc = float(line[2+i+numdpths])
                    mvswc.update({dpth:swc})
                try:
                    Zr = float(line[2+numdpths*2+1])
                except:
                    Zr = float('NaN')
                swp = self.SoilWaterProfile(mdate,
                                            mvswc,
                                            par = self.par,
                                            sol = self.sol,
                                            Zr = Zr)
                self.addprofile(mdate,swp)

    def addprofile(self,mdate,swp):
        """Add a SoilWaterProfile object to self.swdata

        Parameters
        ----------
        mdate : str
            Measurement date as string ('yyyy-ddd')
        swp : SoilWaterProfile object
            Measured soil water profile data on mdate
        """

        self.swdata.update({mdate:swp})

    def customload(self):
        """Override this function to customize loading measured
        vol. soil water content data into SoilWaterProfile objects.
        """

        pass

    def summarize(self):
        """Summarize the series of root zone soil water metrics

        Returns
        -------
        summary : DataFrame
            Summary of series of root zone soil water metrics
        """

        cols = ['Year-DOY','mDr','mDrmax','mfDr','mfDrmax',
                'mSWCr','mSWCrmax','mKs']
        summary = pd.DataFrame(columns = cols)
        summary = summary.set_index('Year-DOY')
        for key in sorted(self.swdata.keys()):
            summary.loc[key,'mDr'] = self.swdata[key].mDr
            summary.loc[key,'mDrmax'] = self.swdata[key].mDrmax
            summary.loc[key,'mfDr'] = self.swdata[key].mfDr
            summary.loc[key,'mfDrmax'] = self.swdata[key].mfDrmax
            summary.loc[key,'mSWCr'] = self.swdata[key].mSWCr
            summary.loc[key,'mSWCrmax'] = self.swdata[key].mSWCrmax
            summary.loc[key,'mKs'] = self.swdata[key].mKs
        return summary

    class SoilWaterProfile:
        """Manage a single soil water content measurement profile

        Attributes
        ----------

        mdate : str
            Year and day of year of the measurement ('yyyy-ddd')
        mvswc : dict
            Measured volumetric soil water content (SWC, cm3/cm3)
            key - Bottom depth of measurement layer as integer (cm)
            value - Measured volumetric SWC as float (cm3/cm3)
        par : pyfao56 Parameters class, optional
            Provides the parameter data (e.g., thetaFC, Zrmax)
            (default = None)
        sol : pyfao56 SoilProfile class, optional
            Provides layered soil profile data (e.g., thetaFC)
            (default = None)
        Zr : float, optional
            Root depth on SWC measurement date (m) (default = NaN)
        mDr : float
            Measured soil water depletion for root zone (mm)
        mDrmax : float
            Measured soil water depletion for maximum root zone (mm)
        mfDr : float
            Measured fractional depletion for root zone (mm/mm)
        mfDrmax : float
            Measured fractional depletion for maximum root zone (mm/mm)
        mSWCr : float
            Measured soil water content for root zone (cm3/cm3)
        mSWCrmax : float
            Measured soil water content for maximum root zone (cm3/cm3)
        mKs : float
            Measured transpiration reduction factor (Ks, 0.0-1.0)

        Methods
        -------
        getZr(model)
            Get Zr on measurement date from model simulation output
        computeDr(negdep=True)
            Compute root zone soil water status metrics from mvswc
        computeKs(model)
            Estimate Ks from measured Dr, TAW, and RAW
        """

        def __init__(self, mdate, mvswc, par = None, sol = None,
                     Zr=float('NaN')):
            """
            Initialize the SoilWaterProfile class attributes.

            Parameters
            ----------
            mdate : str
                Year and day of year of measurement ('yyyy-ddd')
            mvswc : dict
                Measured volumetric soil water content (SWC, cm3/cm3)
                key - Bottom depth of measurement layer as integer (cm)
                value - Measured volumetric SWC as float (cm3/cm3)
            par : pyfao56 Parameters class, optional
                Provides the parameter data (e.g., thetaFC, Zrmax)
                (default = None)
            sol : pyfao56 SoilProfile class, optional
                Provides layered soil profile data (e.g., thetaFC)
                (default = None)
            Zr : float, optional
                Root depth on SWC measurement date (m) (default = NaN)
            """

            self.mdate = mdate
            self.mvswc = mvswc
            self.par = par
            self.sol = sol
            self.Zr = Zr
            self.mDr = float('NaN')
            self.mDrmax = float('NaN')
            self.mfDr = float('NaN')
            self.mfDrmax = float('NaN')
            self.mSWCr = float('NaN')
            self.mSWCrmax = float('NaN')
            self.mKs = float('NaN')

        def __str__(self):
            """Represent the SoilWaterProfile class as a string"""

            s = ('{:8s} '
                 '{:2d} '
                ).format(self.mdate,len(self.mvswc.keys()))
            for key in sorted(self.mvswc.keys()):
                s += '{:3d} '.format(key)
            for key in sorted(self.mvswc.keys()):
                s += '{:5.3f} '.format(self.mvswc[key])
            s += ('{:5.3f} {:8.3f} {:8.3f} {:6.3f} {:7.3f} '
                  '{:5.3f} {:8.3f} {:6.3f}'
                 ).format(self.Zr,self.mDr,self.mDrmax,self.mfDr,
                          self.mfDrmax,self.mSWCr,self.mSWCrmax,
                          self.mKs)
            return s

        def getZr(self, mdl):
            """Get Zr from model simulation on the measurement date

            Parameters
            ----------
            mdl : pyfao56 Model object
                Provides a Model instance with an odata DataFrame
            """

            self.Zr = mdl.odata.loc[self.mdate,'Zr']

        def computeDr(self, negdep = True):
            """Compute root zone soil water status metrics

            Parameters
            ----------
            negdep : boolean, optional
                Allow negative depletion or not (default = True)
            """

            #Root zone is evaluated in 10^-5 meter increments
            #Set root zone depth variables in 10^-5 meter units
            rzmax = int(self.par.Zrmax * 100000.) #10^-5 meters
            rz = int(self.Zr * 100000.) #10^-5 meters

            #Initialize other variables
            swc_dpths = list(self.mvswc.keys())
            if self.sol is not None:
                sol_dpths = list(self.sol.sdata.index.values) #cm
                thetaFC = self.sol.sdata['thetaFC'].to_dict()
                thetaWP = self.sol.sdata['thetaWP'].to_dict()
            elif self.par is not None:
                sol_dpth = int(self.par.Zrmax*100.) #cm
                sol_dpths = [sol_dpth] #cm
                thetaFC = {sol_dpth:self.par.thetaFC}
                thetaWP = {sol_dpth:self.par.thetaWP}
            else:
                raise Exception("No soil profile data available.")
            FCr = 0.
            FCrmax = 0.
            WPr = 0.
            WPrmax = 0.
            SWCr = 0.
            SWCrmax = 0.

            #Iterate the max root zone depth in 10^-5 m increments
            inc_dpth = 0.01 #mm
            for inc in list(range(1, rzmax + 1)):
                #Find soil profile layer depth that contains inc
                sol_dpth = [dpth for (idx, dpth) in enumerate(sol_dpths)
                            if inc <= dpth * 1000][0] #10^-5 meters
                #Find SWC measurement bottom depth that contains inc
                swc_dpth = [dpth for (idx, dpth) in enumerate(swc_dpths)
                            if inc <= dpth * 1000][0] #10^-5 meters
                #Compute incremental values
                FCinc  = thetaFC[sol_dpth] * inc_dpth #mm
                WPinc  = thetaWP[sol_dpth] * inc_dpth #mm
                SWCinc = self.mvswc[swc_dpth] * inc_dpth #mm
                if not negdep and SWCinc > FCinc:
                    SWCinc = FCinc #no negative depletion

                #Accumulate
                FCrmax  += FCinc #mm
                WPrmax  += WPinc #mm
                SWCrmax += SWCinc #mm
                if inc < rz:
                    FCr  += FCinc #mm
                    WPr  += WPinc #mm
                    SWCr += SWCinc #mm

            #Finalize water status metrics
            self.mDr = FCr - SWCr #mm
            self.mDrmax = FCrmax - SWCrmax #mm
            self.mfDr = (FCr - SWCr) / (FCr - WPr) #mm/mm
            self.mfDrmax = (FCrmax-SWCrmax)/(FCrmax-WPrmax) #mm/mm
            self.mSWCr = SWCr / (rz * inc_dpth) #cm3/cm3
            self.mSWCrmax = SWCrmax / (rzmax * inc_dpth) #cm3/cm3

        def computeKs(self, mdl):
            """Estimate Ks from measured Dr, TAW, and RAW

            Parameters
            ----------
            mdl : pyfao56 Model object
                Provides a Model instance with an odata DataFrame
            """

            TAW = mdl.odata.loc[self.mdate,'TAW']
            RAW = mdl.odata.loc[self.mdate,'RAW']
            Ks = (TAW - self.mDr) / (TAW - RAW) #FAO-56 Eq. 84
            self.mKs = sorted([0.0,Ks,1.0])[1]



"""
########################################################################
The parameters.py module contains the Parameters class, which provides
I/O tools for defining input parameters, as required for FAO-56
calculations.

The parameters.py module contains the following:
    Parameters - A class for managing input parameters for FAO-56
        calculations

01/07/2016 Initial Python functions developed by Kelly Thorp
11/04/2021 Finalized updates for inclusion in the pyfao56 Python package
########################################################################
"""

import datetime

class Parameters:
    """A class for managing input parameters for FAO-56 calculations

    Attributes
    ----------
    Kcbini : float
        Kcb Initial (FAO-56 Table 17)
    Kcbmid : float
        Kcb Mid (FAO-56 Table 17)
    Kcbend : float
        Kcb End (FAO-56 Table 17)
    Lini : int
        Length Stage Initial (days) (FAO-56 Table 11)
    Ldev : int
        Length Stage Development (days) (FAO-56 Table 11)
    Lmid : int
        Length Stage Mid (days) (FAO-56 Table 11)
    Lend : int
        Length Stage End (days) (FAO-56 Table 11)
    hini : float
        Plant Height Initial (m)
    hmax : float
        Plant Height Maximum (m) (FAO-56 Table 12)
    thetaFC : float
        Volumetric Soil Water Content, Field Capacity (cm3/cm3)
    thetaWP : float
        Volumetric Soil Water Content, Wilting Point (cm3/cm3)
    theta0 : float
        Volumetric Soil Water Content, Initial (cm3/cm3)
    Zrini : float
        Rooting Depth Initial (m)
    Zrmax : float
        Rooting Depth Maximum (m) (FAO-56 Table 22)
    pbase : float
        Depletion Fraction (p) (FAO-56 Table 22)
    Ze : float
        Depth of surface evaporation layer (m) (FAO-56 Table 19 & p144)
    REW : float
        Total depth Stage 1 evaporation (mm) (FAO-56 Table 19)
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class

    Methods
    -------
    savefile(filepath='pyfao56.par')
        Save the parameter data to a file
    loadfile(filepath='pyfao56.par')
        Load the parameter data from a file
    """

    def __init__(self, Kcbini=0.15, Kcbmid=1.10, Kcbend=0.50, Lini=25,
                 Ldev=50, Lmid=50, Lend=25, hini=0.010, hmax=1.20,
                 thetaFC=0.250, thetaWP=0.100, theta0=0.100, Zrini=0.20,
                 Zrmax=1.40, pbase=0.50, Ze=0.10, REW=8.0, comment=''):
        """Initialize the Parameters class attributes.

        Default parameter values are given below. Users should update
        the parameters with values for their specific crop and field
        conditions based on FAO-56 documentation.

        Parameters
        ----------
        See Parameters class docstring for parameter definitions.
        Kcbini  : float, optional, default = 0.15
        Kcbmid  : float, optional, default = 1.10
        Kcbend  : float, optional, default = 0.50
        Lini    : int  , optional, default = 25
        Ldev    : int  , optional, default = 50
        Lmid    : int  , optional, default = 50
        Lend    : int  , optional, default = 25
        hini    : float, optional, default = 0.010
        hmax    : float, optional, default = 1.20
        thetaFC : float, optional, default = 0.250
        thetaWP : float, optional, default = 0.100
        theta0  : float, optional, default = 0.100
        Zrini   : float, optional, default = 0.20
        Zrmax   : float, optional, default = 1.40
        pbase   : float, optional, default = 0.50
        Ze      : float, optional, default = 0.10
        REW     : float, optional, default = 8.0
        comment : str  , optional, default = ''
        """

        self.Kcbini  = Kcbini
        self.Kcbmid  = Kcbmid
        self.Kcbend  = Kcbend
        self.Lini    = Lini
        self.Ldev    = Ldev
        self.Lmid    = Lmid
        self.Lend    = Lend
        self.hini    = hini
        self.hmax    = hmax
        self.thetaFC = thetaFC
        self.thetaWP = thetaWP
        self.theta0  = theta0
        self.Zrini   = Zrini
        self.Zrmax   = Zrmax
        self.pbase   = pbase
        self.Ze      = Ze
        self.REW     = REW
        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp  = datetime.datetime.now()

    def __str__(self):
        """Represent the Parameter class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        ast='*'*72
        s=('{:s}\n'
           'pyfao56: FAO-56 Evapotranspiration in Python\n'
           'Parameter Data\n'
           'Timestamp: {:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:9.4f} Kcbini, Kcb Initial (FAO-56 Table 17)\n'
           '{:9.4f} Kcbmid, Kcb Mid (FAO-56 Table 17)\n'
           '{:9.4f} Kcbend, Kcb End (FAO-56 Table 17)\n'
           '{:9d} Lini, Length Stage Initial (days) (FAO-56 Table 11)\n'
           '{:9d} Ldev, Length Stage Development (days) '
           '(FAO-56 Table 11)\n'
           '{:9d} Lmid, Length Stage Mid (days) (FAO-56 Table 11)\n'
           '{:9d} Lend, Length Stage End (days) (FAO-56 Table 11)\n'
           '{:9.4f} hini, Plant Height Initial (m)\n'
           '{:9.4f} hmax, Plant Height Maximum (m) (FAO-56 Table 12)\n'
           '{:9.4f} thetaFC, Vol. Soil Water Content, Field Capacity '
           '(cm3/cm3)\n'
           '{:9.4f} thetaWP, Vol. Soil Water Content, Wilting Point '
           '(cm3/cm3)\n'
           '{:9.4f} theta0, Vol. Soil Water Content, Initial '
           '(cm3/cm3)\n'
           '{:9.4f} Zrini, Rooting Depth Initial (m)\n'
           '{:9.4f} Zrmax, Rooting Depth Maximum (m) '
           '(FAO-56 Table 22)\n'
           '{:9.4f} pbase, Depletion Fraction (p) (FAO-56 Table 22)\n'
           '{:9.4f} Ze, Depth of surface evaporation layer (m) '
           '(FAO-56 Table 19 and Page 144)\n'
           '{:9.4f} REW, Total depth Stage 1 evaporation (mm) '
           '(FAO-56 Table 19)\n'
          ).format(ast,timestamp,ast,self.comment,ast,self.Kcbini,
                   self.Kcbmid,self.Kcbend,self.Lini,self.Ldev,
                   self.Lmid,self.Lend,self.hini,self.hmax,self.thetaFC,
                   self.thetaWP,self.theta0,self.Zrini,self.Zrmax,
                   self.pbase,self.Ze,self.REW)
        return s

    def savefile(self,filepath='pyfao56.par'):
        """Save pyfao56 parameters to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.par')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for parameter data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.par'):
        """Load pyfao56 parameters from a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.par')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for parameter data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1] 
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            for line in lines[endast+1:]:
                line = line.strip().split(',')[0].split()
                if line[1].lower() == 'kcbini':
                    self.Kcbini = float(line[0])
                elif line[1].lower() == 'kcbmid':
                    self.Kcbmid = float(line[0])
                elif line[1].lower() == 'kcbend':
                    self.Kcbend = float(line[0])
                elif line[1].lower() == 'lini':
                    self.Lini = int(line[0])
                elif line[1].lower() == 'ldev':
                    self.Ldev = int(line[0])
                elif line[1].lower() == 'lmid':
                    self.Lmid = int(line[0])
                elif line[1].lower() == 'lend':
                    self.Lend = int(line[0])
                elif line[1].lower() == 'hini':
                    self.hini = float(line[0])
                elif line[1].lower() == 'hmax':
                    self.hmax = float(line[0])
                elif line[1].lower() == 'thetafc':
                    self.thetaFC = float(line[0])
                elif line[1].lower() == 'thetawp':
                    self.thetaWP = float(line[0])
                elif line[1].lower() == 'theta0':
                    self.theta0 = float(line[0])
                elif line[1].lower() == 'zrini':
                    self.Zrini = float(line[0])
                elif line[1].lower() == 'zrmax':
                    self.Zrmax = float(line[0])
                elif line[1].lower() == 'pbase':
                    self.pbase = float(line[0])
                elif line[1].lower() == 'ze':
                    self.Ze = float(line[0])
                elif line[1].lower() == 'rew':
                    self.REW = float(line[0])



"""
########################################################################
The model.py module contains the Model class, which defines the
equations for daily soil water balance calculations based on the FAO-56
dual crop coefficient method for evapotranspiration (ET) estimation.

The FAO-56 method is described in the following documentation:
Allen, R. G., Pereira, L. S., Raes, D., Smith, M., 1998.  FAO Irrigation
and Drainage Paper No. 56. Crop Evapotranspiration: Guidelines for
Computing Crop Water Requirements. Food and Agriculture Organization of
the United Nations, Rome Italy.

http://www.fao.org/3/x0490e/x0490e00.htm

The model.py module contains the following:
    Model - A class for managing FAO-56 soil water balance computations.

01/07/2016 Initial Python functions developed by Kelly Thorp
11/04/2021 Finalized updates for inclusion in the pyfao56 Python package
10/27/2022 Incorporated Fort Collins ARS stratified soil layers approach
11/30/2022 Incorporated Fort Collins ARS water balance approach
08/17/2023 Improved logic for case of missing rhmin data
10/31/2023 Added AquaCrop Ks option
11/01/2023 Added reports of the cumulative seasonal water balance
########################################################################
"""

import pandas as pd
import datetime
import math

class Model:
    """A class for managing FAO-56 soil water balance computations.

    Manages computations based on FAO-56 methods for evapotranspiration
    and soil water balance calculations (Allen et al., 1998).

    Attributes
    ----------
    startDate : datetime
        Simulation start date in datetime format
    endDate : datetime
        Simulation end date in datetime format
    par : pyfao56 Parameters class
        Provides the parameter data for simulations
    wth : pyfao56 Weather class
        Provides the weather data for simulations
    irr : pyfao56 Irrigation class, optional
        Provides the irrigation data for simulations
        (default = None)
    sol : pyfao56 SoilProfile class, optional
        Provides data for modeling with stratified soil layers
        (default = None)
    upd : pyfao56 Update class, optional
        Provides data and methods for state variable updating
        (default = None)
    cons_p : boolean, optional
        If False, p follows FAO-56; if True, p is constant (=pbase)
        (default = False)
    aq_Ks : boolean, optional
        If False, Ks follows FAO-56; if True, Ks via AquaCrop equation
        (default = False)
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    ModelState : class
        Contains parameters and model states for a single timestep
    cnames : list
        Column names for odata
    odata : DataFrame
        Model output data as float
        index - Year and day of year as string ('yyyy-ddd')
        columns - ['Year','DOY','DOW','Date','ETref','tKcb','Kcb','h',
                   'Kcmax','fc','fw','few','De','Kr','Ke','E','DPe',
                   'Kc','ETc','TAW','TAWrmax','TAWb','Zr','p','RAW',
                   'Ks','Kcadj','ETcadj','T','DP','Dinc','Dr','fDr',
                   'Drmax','fDrmax','Db','fDb','Irrig','Rain','Year',
                   'DOY','DOW','Date']
            Year    - 4-digit year (yyyy)
            DOY     - Day of year (ddd)
            DOW     - Day of week
            Date    - Month/Day/Year (mm/dd/yy)
            ETref   - Daily reference evapotranspiration (mm)
            tKcb    - Basal crop coefficient, trapezoidal from FAO-56
            Kcb     - Basal crop coefficient, considering updates
            h       - Plant height (m)
            Kcmax   - Upper limit crop coefficient, FAO-56 Eq. 72
            fc      - Canopy cover fraction, FAO-56 Eq. 76
            fw      - Fraction soil surface wetted, FAO-56 Table 20
            few     - Exposed & wetted soil fraction, FAO-56 Eq. 75
            De      - Cumulative depth of evaporation, FAO-56 Eqs. 77&78
            Kr      - Evaporation reduction coefficient, FAO-56 Eq. 74
            Ke      - Evaporation coefficient, FAO-56 Eq. 71
            E       - Soil water evaporation (mm), FAO-56 Eq. 69
            DPe     - Percolation under exposed soil (mm), FAO-56 Eq. 79
            Kc      - Crop coefficient, FAO-56 Eq. 69
            ETc     - Non-stressed crop ET (mm), FAO-56 Eq. 69
            TAW     - Total available water (mm), FAO-56 Eq. 82
            TAWrmax - Total available water for max root depth (mm)
            TAWb    - Total available water in bottom layer (mm)
            Zr      - Root depth (m), FAO-56 page 279
            p       - Fraction depleted TAW, FAO-56 p162 and Table 22
            RAW     - Readily available water (mm), FAO-56 Equation 83
            Ks      - Transpiration reduction factor, FAO-56 Eq. 84
            Kcadj   - Adjusted crop coefficient, FA0-56 Eq. 80
            ETcadj  - Adjusted crop ET (mm), FAO-56 Eq. 80
            T       - Adjusted crop transpiration (mm)
            DP      - Deep percolation (mm), FAO-56 Eq. 88
            Dinc    - Depletion increment due to root growth (mm)
            Dr      - Soil water depletion (mm), FAO-56 Eqs. 85 & 86
            fDr     - Fractional root zone soil water depletion (mm/mm)
            Drmax   - Soil water depletion for max root depth (mm)
            fDrmax  - Fractional depletion for max root depth (mm/mm)
            Db      - Soil water depletion in the bottom layer (mm)
            fDb     - Fractional depletion in the bottom layer (mm/mm)
            Irrig   - Depth of irrigation (mm)
            Rain    - Depth of precipitation (mm)
            Year    - 4-digit year (yyyy)
            DOY     - Day of year (ddd)
            DOW     - Day of week
            Date    - Month/Day/Year (mm/dd/yy)
    swbdata : dict
        Container for cumulative seasonal water balance data
        keys - ['ETref','ETc','ETcadj','E','T','DP','Irrig','Rain',
                'Dr_ini','Dr_end','Drmax_ini','Drmax_end']
        value - Cumulative water balance data in mm

    Methods
    -------
    savefile(filepath='pyfao56.out')
        Save pyfao56 output data to a file
    savesums(filepath='pyfao56.sum')
        Save seasonal water balance data to a file
    run()
        Conduct the FAO-56 calculations from startDate to endDate
    """

    def __init__(self, start, end, par, wth, irr=None, sol=None,
                 upd=None, cons_p=False, aq_Ks=False, comment=''):
        """Initialize the Model class attributes.

        Parameters
        ----------
        start : str
            Simulation start year and doy ('yyyy-ddd')
        end : str
            Simulation end year and doy ('yyyy-ddd')
        par : pyfao56 Parameters object
            Provides the parameter data for simulations
        wth : pyfao56 Weather object
            Provides the weather data for simulations
        irr : pyfao56 Irrigation object, optional
            Provides the irrigation data for simulations
            (default = None)
        sol : pyfao56 SoilProfile object, optional
            Provides data for modeling with stratified soil layers
            (default = None)
        upd : pyfao56 Update object, optional
            Provides data and methods for state variable updating
            (default = None)
        cons_p : boolean, optional
            If False, p follows FAO-56; if True, p is constant (=pbase)
            (default = False)
        aq_Ks : boolean, optional
            If False, Ks follows FAO-56; if True, Ks via AquaCrop Eqn
            (default = False)
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.startDate = datetime.datetime.strptime(start, '%Y-%j')
        self.endDate   = datetime.datetime.strptime(end, '%Y-%j')
        self.par = par
        self.wth = wth
        self.irr = irr
        self.sol = sol
        self.upd = upd
        self.cons_p = cons_p
        self.aq_Ks = aq_Ks
        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.cnames = ['Year','DOY','DOW','Date','ETref','tKcb','Kcb',
                       'h','Kcmax','fc','fw','few','De','Kr','Ke','E',
                       'DPe','Kc','ETc','TAW','TAWrmax','TAWb','Zr','p',
                       'RAW','Ks','Kcadj','ETcadj','T','DP','Dinc','Dr',
                       'fDr','Drmax','fDrmax','Db','fDb','Irrig','Rain',
                       'Year','DOY','DOW','Date']
        self.odata = pd.DataFrame(columns=self.cnames)

    def __str__(self):
        """Represent the Model class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        sdate = self.startDate.strftime('%m/%d/%Y')
        edate = self.endDate.strftime('%m/%d/%Y')
        if self.sol is None:
            solmthd = 'D - Default FAO-56 homogenous soil bucket ' \
                      'approach'
        else:
            solmthd = 'L - Fort Collins ARS stratified soil layers ' \
                      'approach'
        fmts = {'Year':'{:4s}'.format,'DOY':'{:3s}'.format,
                'DOW':'{:3s}'.format,'Date':'{:8s}'.format,
                'ETref':'{:6.3f}'.format,'tKcb':'{:5.3f}'.format,
                'Kcb':'{:5.3f}'.format,'h':'{:5.3f}'.format,
                'Kcmax':'{:5.3f}'.format,'fc':'{:5.3f}'.format,
                'fw':'{:5.3f}'.format,'few':'{:5.3f}'.format,
                'De':'{:7.3f}'.format,'Kr':'{:5.3f}'.format,
                'Ke':'{:5.3f}'.format,'E':'{:6.3f}'.format,
                'DPe':'{:7.3f}'.format,'Kc':'{:5.3f}'.format,
                'ETc':'{:6.3f}'.format,'TAW':'{:7.3f}'.format,
                'TAWrmax':'{:7.3f}'.format,'TAWb':'{:7.3f}'.format,
                'Zr':'{:5.3f}'.format,'p':'{:5.3f}'.format,
                'RAW':'{:7.3f}'.format,'Ks':'{:5.3f}'.format,
                'Kcadj':'{:5.3f}'.format,'ETcadj':'{:6.3f}'.format,
                'T':'{:6.3f}'.format,'DP':'{:7.3f}'.format,
                'Dinc':'{:7.3f}'.format,'Dr':'{:7.3f}'.format,
                'fDr':'{:7.3f}'.format,'Drmax':'{:7.3f}'.format,
                'fDrmax':'{:7.3f}'.format,'Db':'{:7.3f}'.format,
                'fDb':'{:7.3f}'.format,'Irrig':'{:7.3f}'.format,
                'Rain':'{:7.3f}'.format}
        ast='*'*72
        s = ('{:s}\n'
             'pyfao56: FAO-56 Evapotranspiration in Python\n'
             'Output Data\n'
             'Timestamp: {:s}\n'
             'Simulation start date: {:s}\n'
             'Simulation end date: {:s}\n'
             'Soil method: {:s}\n'
             '{:s}\n'
             '{:s}\n'
             '{:s}\n'
             'Year-DOY  Year  DOY  DOW      Date  ETref  tKcb   Kcb'
             '     h Kcmax    fc    fw   few      De    Kr    Ke      E'
             '     DPe    Kc    ETc     TAW TAWrmax    TAWb    Zr     p'
             '     RAW    Ks Kcadj ETcadj      T      DP    Dinc'
             '      Dr     fDr   Drmax  fDrmax      Db     fDb'
             '   Irrig    Rain  Year  DOY  DOW      Date\n'
             ).format(ast,
                      timestamp,
                      sdate,
                      edate,
                      solmthd,
                      ast,
                      self.comment,
                      ast)
        if not self.odata.empty:
            s += self.odata.to_string(header=False,formatters=fmts)
        return s

    def savefile(self,filepath='pyfao56.out'):
        """Save pyfao56 output data to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.out')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for output data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def savesums(self, filepath='pyfao56.sum'):
        """Save a summary file with cumulative water balance values.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.sum')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        sdate = self.startDate.strftime('%m/%d/%Y')
        edate = self.endDate.strftime('%m/%d/%Y')

        ast = '*'*72
        s = ('{:s}\n'
            'pyfao56: FAO-56 Evapotranspiration in Python\n'
            'Seasonal Water Balance Summary\n'
            'Timestamp: {:s}\n'
            'Simulation start date: {:s}\n'
            'Simulation end date: {:s}\n'
            'All values expressed in mm.\n'
            '{:s}\n'
            '{:s}\n'
            '{:s}\n'
            ).format(ast,timestamp,sdate,edate,ast,self.comment,ast)
        if not self.odata.empty:
            keys = ['ETref','ETc','ETcadj','E','T','DP','Irrig','Rain',
                    'Dr_ini','Dr_end','Drmax_ini','Drmax_end']
            for key in keys:
                s += '{:8.3f} : {:s}\n'.format(self.swbdata[key],key)

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for summary data is not found.')
        else:
            f.write(s)
            f.close()

    class ModelState:
        """Contain parameters and states for a single timestep."""

        pass

    def run(self):
        """Initialize model, conduct simulations, update self.odata"""

        tcurrent = self.startDate
        tdelta = datetime.timedelta(days=1)

        #Initialize model state
        io = self.ModelState()
        io.i = 0
        io.Kcbini  = self.par.Kcbini
        io.Kcbmid  = self.par.Kcbmid
        io.Kcbend  = self.par.Kcbend
        io.Lini    = self.par.Lini
        io.Ldev    = self.par.Ldev
        io.Lmid    = self.par.Lmid
        io.Lend    = self.par.Lend
        io.hini    = self.par.hini
        io.hmax    = self.par.hmax
        io.thetaFC = self.par.thetaFC
        io.thetaWP = self.par.thetaWP
        io.theta0  = self.par.theta0
        io.Zrini   = self.par.Zrini
        io.Zrmax   = self.par.Zrmax
        io.pbase   = self.par.pbase
        io.Ze      = self.par.Ze
        io.REW     = self.par.REW
        #Total evaporable water (TEW, mm) - FAO-56 Equation 73
        io.TEW = 1000. * (io.thetaFC - 0.50 * io.thetaWP) * io.Ze
        #Initial depth of evaporation (De, mm) - FAO-56 page 153
        io.De = 1000. * (io.thetaFC - 0.50 * io.thetaWP) * io.Ze
        if self.sol is None:
            io.solmthd = 'D' #Default homogeneous soil from Parameters
            #Initial root zone depletion (Dr, mm) - FAO-56 Equation 87
            io.Dr = 1000. * (io.thetaFC - io.theta0) * io.Zrini
            #Initial soil depletion for max root depth (Drmax, mm)
            io.Drmax = 1000. * (io.thetaFC - io.theta0) * io.Zrmax
            #Initial root zone total available water (TAW, mm)
            io.TAW = 1000. * (io.thetaFC - io.thetaWP) * io.Zrini
            #By default, FAO-56 doesn't consider the following variables
            io.TAWrmax = -99.999
            io.Db = -99.999
            io.TAWb = -99.999
        else:
            io.solmthd = 'L' #Layered soil profile from SoilProfile
            io.lyr_dpths = list(self.sol.sdata.index)
            io.lyr_thFC  = list(self.sol.sdata['thetaFC'])
            io.lyr_thWP  = list(self.sol.sdata['thetaWP'])
            io.lyr_th0   = list(self.sol.sdata['theta0'])
            io.Dr = 0.
            io.Drmax = 0.
            io.TAW = 0.
            io.TAWrmax = 0.
            #Iterate down the soil profile in 1 mm increments
            for dpthmm in list(range(1, (io.lyr_dpths[-1] * 10 + 1))):
                #Find soil layer index that contains dpthmm
                lyr_idx = [idx for (idx, dpth) in
                          enumerate(io.lyr_dpths) if dpthmm<=dpth*10][0]
                #Initial root zone depletion (Dr, mm)
                if dpthmm <= io.Zrini * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_th0[lyr_idx])
                    io.Dr += diff #mm
                #Initial depletion for max root depth (Drmax, mm)
                if dpthmm <= io.Zrmax * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_th0[lyr_idx])
                    io.Drmax += diff #mm
                #Initial root zone total available water (TAW, mm)
                if dpthmm <= io.Zrini * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_thWP[lyr_idx])
                    io.TAW += diff #mm
                #Total available water for max root depth (TAWrmax, mm)
                if dpthmm <= io.Zrmax * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_thWP[lyr_idx])
                    io.TAWrmax += diff #mm
            #Initial depletion in the bottom layer (Db, mm)
            io.Db = io.Drmax - io.Dr
            #Initial total available water in bottom layer (TAWb, mm)
            io.TAWb = io.TAWrmax - io.TAW
        io.h = io.hini
        io.Zr = io.Zrini
        io.fw = 1.0
        io.wndht = self.wth.wndht
        io.rfcrp = self.wth.rfcrp
        io.cons_p = self.cons_p
        io.aq_Ks = self.aq_Ks
        self.odata = pd.DataFrame(columns=self.cnames)

        while tcurrent <= self.endDate:
            mykey = tcurrent.strftime('%Y-%j')

            #Update ModelState object
            io.ETref = self.wth.wdata.loc[mykey,'ETref']
            if math.isnan(io.ETref):
                io.ETref = self.wth.compute_etref(mykey)
            io.rain = self.wth.wdata.loc[mykey,'Rain']
            io.wndsp = self.wth.wdata.loc[mykey,'Wndsp']
            if math.isnan(io.wndsp):
                io.wndsp = 2.0
            io.rhmin = self.wth.wdata.loc[mykey,'RHmin']
            if math.isnan(io.rhmin):
                tmax = self.wth.wdata.loc[mykey,'Tmax']
                tmin = self.wth.wdata.loc[mykey,'Tmin']
                tdew = self.wth.wdata.loc[mykey,'Tdew']
                if math.isnan(tdew):
                    tdew = tmin
                #ASCE (2005) Eqs. 7 and 8
                emax = 0.6108*math.exp((17.27*tmax)/
                                       (tmax+237.3))
                ea   = 0.6108*math.exp((17.27*tdew)/
                                       (tdew+237.3))
                io.rhmin = ea/emax*100.
            if math.isnan(io.rhmin):
                io.rhmin = 45.
            io.idep = 0.0
            if self.irr is not None:
                if mykey in self.irr.idata.index:
                    io.idep = self.irr.idata.loc[mykey,'Depth']
                    io.fw = self.irr.idata.loc[mykey,'fw']

            #Obtain updates for Kcb, h, and fc, if available
            io.updKcb = float('NaN')
            io.updh = float('NaN')
            io.updfc = float('NaN')
            if self.upd is not None:
                io.updKcb = self.upd.getdata(mykey,'Kcb')
                io.updh = self.upd.getdata(mykey,'h')
                io.updfc = self.upd.getdata(mykey,'fc')

            #Advance timestep
            self._advance(io)

            #Append results to self.odata
            year = tcurrent.strftime('%Y')
            doy = tcurrent.strftime('%j') #Day of Year
            dow = tcurrent.strftime('%a') #Day of Week
            dat = tcurrent.strftime('%m/%d/%y') #Date mm/dd/yy
            data = [year, doy, dow, dat, io.ETref, io.tKcb, io.Kcb,
                    io.h, io.Kcmax, io.fc, io.fw, io.few, io.De, io.Kr,
                    io.Ke, io.E, io.DPe, io.Kc, io.ETc, io.TAW,
                    io.TAWrmax, io.TAWb, io.Zr, io.p, io.RAW, io.Ks,
                    io.Kcadj, io.ETcadj, io.T, io.DP, io.Dinc, io.Dr,
                    io.fDr, io.Drmax, io.fDrmax, io.Db, io.fDb, io.idep,
                    io.rain, year, doy, dow, dat]
            self.odata.loc[mykey] = data

            tcurrent = tcurrent + tdelta
            io.i+=1

        #Save seasonal water balance data to self.swbdata dictionary
        sdoy = self.startDate.strftime("%Y-%j")
        edoy = self.endDate.strftime("%Y-%j")
        self.swbdata = {'ETref'    :sum(self.odata['ETref']),
                        'ETc'      :sum(self.odata['ETc']),
                        'ETcadj'   :sum(self.odata['ETcadj']),
                        'E'        :sum(self.odata['E']),
                        'T'        :sum(self.odata['T']),
                        'DP'       :sum(self.odata['DP']),
                        'Irrig'    :sum(self.odata['Irrig']),
                        'Rain'     :sum(self.odata['Rain']),
                        'Dr_ini'   :self.odata.loc[sdoy,'Dr'],
                        'Dr_end'   :self.odata.loc[edoy,'Dr'],
                        'Drmax_ini':self.odata.loc[sdoy,'Drmax'],
                        'Drmax_end':self.odata.loc[edoy,'Drmax']}

    def _advance(self, io):
        """Advance the model by one daily timestep.

        Parameters
        ----------
        io : ModelState object
        """

        #Basal crop coefficient (Kcb)
        #From FAO-56 Tables 11 and 17
        s1 = io.Lini
        s2 = s1 + io.Ldev
        s3 = s2 + io.Lmid
        s4 = s3 + io.Lend
        if 0<=io.i<=s1:
            io.tKcb = io.Kcbini
            io.Kcb = io.Kcbini
        elif s1<io.i<=s2:
            io.tKcb += (io.Kcbmid-io.Kcbini)/(s2-s1)
            io.Kcb += (io.Kcbmid-io.Kcbini)/(s2-s1)
        elif s2<io.i<=s3:
            io.tKcb = io.Kcbmid
            io.Kcb = io.Kcbmid
        elif s3<io.i<=s4:
            io.tKcb += (io.Kcbmid-io.Kcbend)/(s3-s4)
            io.Kcb += (io.Kcbmid-io.Kcbend)/(s3-s4)
        elif s4<io.i:
            io.tKcb = io.Kcbend
            io.Kcb = io.Kcbend
        #Overwrite Kcb if updates are available
        if io.updKcb > 0: io.Kcb = io.updKcb

        #Plant height (h, m)
        io.h = max([io.hini+(io.hmax-io.hini)*(io.Kcb-io.Kcbini)/
                    (io.Kcbmid-io.Kcbini),0.001,io.h])
        #Overwrite h if updates are available
        if io.updh > 0: io.h = io.updh

        #Root depth (Zr, m) - FAO-56 page 279
        io.Zr = max([io.Zrini + (io.Zrmax-io.Zrini)*(io.tKcb-io.Kcbini)/
                     (io.Kcbmid-io.Kcbini),0.001,io.Zr])

        #Upper limit crop coefficient (Kcmax) - FAO-56 Eq. 72
        u2 = io.wndsp * (4.87/math.log(67.8*io.wndht-5.42))
        u2 = sorted([1.0,u2,6.0])[1]
        rhmin = sorted([20.0,io.rhmin,80.])[1]
        if io.rfcrp == 'S':
            io.Kcmax = max([1.2+(0.04*(u2-2.0)-0.004*(rhmin-45.0))*
                            (io.h/3.0)**.3, io.Kcb+0.05])
        elif io.rfcrp == 'T':
            io.Kcmax = max([1.0, io.Kcb + 0.05])

        #Canopy cover fraction (fc, 0.0-0.99) - FAO-56 Eq. 76
        io.fc = sorted([0.0,((io.Kcb-io.Kcbini)/(io.Kcmax-io.Kcbini))**
                        (1.0+0.5*io.h),0.99])[1]
        #Overwrite fc if updates are available
        if io.updfc > 0: io.fc = io.updfc

        #Fraction soil surface wetted (fw) - FAO-56 Table 20, page 149
        if io.idep > 0.0 and io.rain > 0.0:
            pass #fw=fw input
        elif io.idep > 0.0 and io.rain <= 0.0:
            pass #fw=fw input
        elif io.idep <= 0.0 and io.rain >= 3.0:
            io.fw = 1.0
        else:
            pass #fw = previous fw

        #Exposed & wetted soil fraction (few, 0.01-1.0) - FAO-56 Eq. 75
        io.few = sorted([0.01,min([1.0-io.fc, io.fw]),1.0])[1]

        #Evaporation reduction coefficient (Kr, 0-1) - FAO-56 Eq. 74
        io.Kr = sorted([0.0,(io.TEW-io.De)/(io.TEW-io.REW),1.0])[1]

        #Evaporation coefficient (Ke) - FAO-56 Eq. 71
        io.Ke = min([io.Kr*(io.Kcmax-io.Kcb), io.few*io.Kcmax])

        #Soil water evaporation (E, mm) - FAO-56 Eq. 69
        io.E = io.Ke * io.ETref

        #Deep percolation under exposed soil (DPe, mm) - FAO-56 Eq. 79
        runoff = 0.0
        io.DPe = max([io.rain - runoff + io.idep/io.fw - io.De,0.0])

        #Cumulative depth of evaporation (De, mm) - FAO-56 Eqs. 77 & 78
        De = io.De-(io.rain-runoff)-io.idep/io.fw+io.E/io.few+io.DPe
        io.De = sorted([0.0,De,io.TEW])[1]

        #Crop coefficient (Kc) - FAO-56 Eq. 69
        io.Kc = io.Ke + io.Kcb

        #Non-stressed crop evapotranspiration (ETc, mm) - FAO-56 Eq. 69
        io.ETc = io.Kc * io.ETref

        if io.solmthd == 'D':
            # Total available water (TAW, mm) - FAO-56 Eq. 82
            io.TAW = 1000.0 * (io.thetaFC - io.thetaWP) * io.Zr
        elif io.solmthd == 'L':
            io.TAW = 0.
            #Iterate down the soil profile in 1 mm increments
            for dpthmm in list(range(1, (io.lyr_dpths[-1] * 10 + 1))):
                #Find soil layer index that contains dpthmm
                lyr_idx = [idx for (idx, dpth) in
                          enumerate(io.lyr_dpths) if dpthmm<=dpth*10][0]
                #Total available water (TAW, mm)
                if dpthmm <= io.Zr * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_thWP[lyr_idx])
                    io.TAW += diff #mm
            #Total available water in the bottom layer (TAWb, mm)
            io.TAWb_prev = io.TAWb
            io.TAWb = io.TAWrmax - io.TAW

        #Fraction depleted TAW (p, 0.1-0.8) - FAO-56 p162 and Table 22
        if io.cons_p is True:
            io.p = io.pbase
        else:
            io.p = sorted([0.1,io.pbase+0.04*(5.0-io.ETc),0.8])[1]

        #Readily available water (RAW, mm) - FAO-56 Equation 83
        io.RAW = io.p * io.TAW

        #Transpiration reduction factor (Ks, 0.0-1.0)
        if io.aq_Ks is True:
            #Ks method from AquaCrop
            rSWD = io.Dr/io.TAW
            Drel = (rSWD-io.p)/(1.0-io.p)
            sf = 1.5
            aqKs = 1.0-(math.exp(sf*Drel)-1.0)/(math.exp(sf)-1.0)
            io.Ks = sorted([0.0, aqKs, 1.0])[1]
        else:
            #FAO-56 Eq. 84
            io.Ks = sorted([0.0,(io.TAW-io.Dr)/(io.TAW-io.RAW),1.0])[1]

        #Adjusted crop coefficient (Kcadj) - FAO-56 Eq. 80
        io.Kcadj = io.Ks * io.Kcb + io.Ke

        #Adjusted crop evapotranspiration (ETcadj, mm) - FAO-56 Eq. 80
        io.ETcadj = io.Kcadj * io.ETref

        #Adjusted crop transpiration (T, mm)
        io.T = (io.Ks * io.Kcb) * io.ETref

        #Water balance methods
        if io.solmthd == 'D':
            #Deep percolation (DP, mm) - FAO-56 Eq. 88
            #Boundary layer is considered at the root zone depth (Zr)
            io.DP = max([io.rain-runoff+io.idep-io.ETcadj-io.Dr,0.0])

            #Root zone soil water depletion (Dr,mm) - FAO-56 Eqs.85 & 86
            Dr = io.Dr - (io.rain-runoff) - io.idep + io.ETcadj + io.DP
            io.Dr = sorted([0.0, Dr, io.TAW])[1]

            #Root zone soil water depletion fraction (fDr, mm/mm)
            io.fDr = 1.0 - ((io.TAW - io.Dr) / io.TAW)

            #By default, FAO-56 doesn't consider the following variables
            io.Dinc = -99.999
            io.Drmax = -99.999
            io.fDrmax = -99.999
            io.Db = -99.999
            io.fDb = -99.999

        elif io.solmthd == 'L':
            #Deep percolation (DP, mm)
            #Boundary layer is at the max root depth (Zrmax)
            io.DP = max([io.rain-runoff+io.idep-io.ETcadj-io.Drmax,0.0])

            #Depletion increment due to root growth (Dinc, mm)
            #Computed from Db based on the incremental change in TAWb
            if io.TAWb_prev > 0.0:
                io.Dinc = io.Db * (1.0 - (io.TAWb / io.TAWb_prev))
            else: #handle zero divide issue
                io.Dinc = 0.0

            #Root zone soil water depletion (Dr, mm)
            Dr = io.Dr - (io.rain-runoff)-io.idep+io.ETcadj+io.Dinc
            io.Dr = sorted([0.0, Dr, io.TAW])[1]

            #Root zone soil water depletion fraction (fDr, mm/mm)
            io.fDr = 1.0 - ((io.TAW - io.Dr) / io.TAW)

            #Soil water depletion at max root depth (Drmax, mm)
            Drmax = io.Drmax - (io.rain-runoff)-io.idep+io.ETcadj+io.DP
            io.Drmax = sorted([0.0, Drmax, io.TAWrmax])[1]

            #Soil water depletion fraction at Zrmax (fDrmax, mm/mm)
            io.fDrmax = 1.0 - ((io.TAWrmax - io.Drmax) / io.TAWrmax)

            #Soil water depletion in the bottom layer (Db, mm)
            Db = io.Drmax - io.Dr
            io.Db = sorted([0.0, Db, io.TAWb])[1]

            #Bottom layer soil water depletion fraction (fDb, mm/mm)
            if io.TAWb > 0.0:
                io.fDb = 1.0 - ((io.TAWb - io.Db) / io.TAWb)
            else:
                io.fDb = 0.0
